
(******************************************************************************)
(** Types from .mli ***********************************************************)
(******************************************************************************)

(* types declared in a module type T need to be defined in any module
   that implements T.  In this case, the module type Algebra (defined in
   algebra.mli) declares a module type Addable, so the module Algebra (defined
   in this file) must define a module type called Addable as well.  In practice,
   this means types defined in a .mli file must often be copied into the .ml
   file.

   There are also modules and values declared in the .mli file; these
   implementations are below (in the "Implementations" section)
*)

module type AddableCore = sig
  type t
  val (+)   : t -> t -> t
  val (~-)  : t -> t
  val zero : t
  val (=)  : t -> t -> bool
end

module type AddableExtension = sig
  type t
  val (-)    : t -> t -> t
  val (<>)   : t -> t -> bool
end

module type Addable = sig
  type t
  include AddableCore      with type t := t
  include AddableExtension with type t := t
end

module type MultipliableCore = sig
  type t
  include AddableCore with type t := t
  val ( * ) : t -> t -> t
  val one   : t
end

module type MultipliableExtension = sig
  include AddableExtension
  val from_int : int -> t
end

module type Multipliable = sig
  type t
  include MultipliableCore      with type t := t
  include MultipliableExtension with type t := t
end

module type OrderedCore = sig
  include MultipliableCore

  val isPositive : t -> bool
end

module type OrderedExtension = sig
  include MultipliableExtension

  include Set.OrderedType with type t := t
  include Map.OrderedType with type t := t

  val (<)  : t -> t -> bool
  val (>)  : t -> t -> bool
  val (<=) : t -> t -> bool
  val (>=) : t -> t -> bool
end

module type Ordered = sig
  include OrderedCore
  include OrderedExtension with type t := t
end

module type DividableCore = sig
  include MultipliableCore

  val inv : t -> t
end

module type DividableExtension = sig
  include MultipliableExtension

  val (/) : t -> t -> t
end

module type Dividable = sig
  include DividableCore
  include DividableExtension with type t := t
end

module type NumberCore = sig
  type t

  include DividableCore with type t := t
  include OrderedCore   with type t := t

  val to_float : t -> float
end

module type NumberExtension = sig
  include DividableExtension
  include OrderedExtension   with type t := t
end

module type Number = sig
  include NumberCore
  include NumberExtension with type t := t
end

(******************************************************************************)
(** Implementations                                                           *)
(******************************************************************************)

module ExtendAddable (C : AddableCore)
  : Addable with type t := C.t =
struct
  include C

  let (-)  a b = a + (-b)
  let (<>) a b = not (a = b)
end

module ExtendMultipliable (C : MultipliableCore)
  : Multipliable with type t := C.t =
struct
  include ExtendAddable(C)
  include C

  let rec from_int n =
    let open Pervasives in
    if n = 0 then zero
    else if n < 0 then C.(~-) (from_int (-n))
    else C.(+) C.one (from_int (n-1))
end

module ExtendOrdered (C : OrderedCore)
 : Ordered with type t := C.t =
struct
  include C
  include ExtendMultipliable(C)

  let (>)  a b = isPositive (a - b)
  let (<)  a b = isPositive (b - a)
  let (>=) a b = a > b || a = b
  let (<=) a b = a < b || a = b

  let compare a b =
    if a > b then 1
    else if a = b then 0
    else -1
end

module ExtendDividable (C : DividableCore)
  : Dividable with type t := C.t =
struct
  include C
  include ExtendMultipliable(C)

  let (/) a b = a * (inv b)
end

module ExtendNumber (C : NumberCore)
  : Number with type t := C.t =
struct
  include C
  include ExtendOrdered(C)
  include ExtendDividable(C)
end


