open Algebra

(******************************************************************************)
(* Implementing the Ordered interface with ints *******************************)
(******************************************************************************)

module Int = struct
  (** We define the core module, which tells how to perform basic operations on
      ints. *)
  module Core = struct
    type t = int

    let (=) a b = a = b
    let (+) a b = a + b
    let (~-) a  = -a
    let zero    = 0

    let ( * ) a b = a * b
    let one = 1

    let isPositive a = a > 0
  end

  (* alternate implementation: inherit the definitions directly from Pervasives
  module Core = struct
    type t = int
    include Pervasives
    let zero = 0
    let one  = 1
    let isPositive a = a > 0
  end
  *)

  (** We include all the definitions from the Core modulue into the Ints module
      itself.  This allows us to write Ints.zero instead of Ints.Core.zero. This
      also makes Ints itself an implementation of the AddableCore interface *)
  include Core

  (** We then apply the ExtendedOrdered functor to the Core module, which
      produces definitions for all of the ExtendedAddable functions.  By
      including this module into the Ints module, we gain implementations of
      all of the AddableExtension functions.

      This makes the Int module an implementation of the AddableExtension
      interface, and also an implementation of the full Addable interface
    *)
  include ExtendOrdered(Core)
end

(******************************************************************************)
(** Other implementations (these are similar to Int) **************************)
(******************************************************************************)

(* floats *)

module Float = struct
  module Core = struct
    type t = float

    let (=) a b = (a = b)
    let (+) a b = a +. b
    let (~-) a  = -.a
    let zero    = 0.

    let ( * ) = ( *. )
    let one   = 1.

    let isPositive a = a > 0.

    let inv a = 1. /. a
    let to_float a = a
  end

  include Core
  include ExtendNumber(Core)
end

(* big integers.

   Note: to compile this, you need to include the ocaml nums library.  Be sure
   that the attached .cs3110 file is in your current directory if compiling with
   the cs3110 tools.

   See the OCaml BigInt module for a list of big_int functions
*)
module BigInt = struct
  module Core = struct
    open Big_int
    type t = big_int

    let (+)   = add_big_int
    let (~-)  = minus_big_int
    let ( * ) = mult_big_int
    let (=)   = eq_big_int

    let one  = unit_big_int
    let zero = zero_big_int

    let isPositive n = (sign_big_int n) > 0
    let to_float     = float_of_big_int
  end

  include Core
  include ExtendOrdered(Core)
end

(******************************************************************************)
(** The Fraction functor ******************************************************)
(******************************************************************************)

(** We could implement fractions as pairs of ints, or as pairs of bigints.  By
    parameterizing the Fraction module on the Multipliable interface, we can get
    both implementations: Fraction(Int) will store pairs of ints, while
    Fraction(BigInt) will store pairs of big integers.  We can get fancy and
    implement rational functions as Fraction(Polynomial); no need to implement
    a polynomial division algorithm.  *)

(* see .mli *)
module type Fraction = sig
  module Numerator : Multipliable
  include Dividable

  val from_numerator : Numerator.t -> t
  val to_pair        : t -> Numerator.t * Numerator.t
end

(** Fraction functor implementation *)
module Fraction (M : Multipliable) = struct
  module Core = struct
    type t = M.t * M.t

    let (=)   (n1,d1) (n2,d2) = M.(n1*d2 = n2*d1)
    let (+)   (n1,d1) (n2,d2) = M.(n1*d2 + n2*d1, d1*d2)
    let (~-)  (n1,d1)         = M.(-n1, d1)
    let zero                  = M.(zero, one)
    let one                   = M.(one,  one)
    let ( * ) (n1,d1) (n2,d2) = M.(n1*n2, d1*d2)
    let inv (n, d) = if M.(n = zero) then failwith "Division by zero!" else (d,n)

  end

  include Core
  include ExtendDividable(Core)

  module Numerator = M

  let to_pair f = f
  let from_numerator n = (n, M.one)
end

(** In implementing the Fraction module, we had a choice: we could require the
    numerator and denominator to be Ordered, in which case we can implement
    Fractions(M).isPositive in terms of M.isPositive.  But this prevents us
    from forming fractions of non-ordered things.  Luckily, we can extend the
    Fraction module to create an OrdFraction module; we simply include Fraction.
    *)

(* see .mli *)
module type OrdFraction = sig
  include Fraction
  include Ordered with type t := t
end

module OrdFraction (M : Ordered) = struct
  module OCore = struct
    module FC = Fraction(M)
    include FC.Core
    let isPositive (n,d) = Pervasives.((M.isPositive n) = (M.isPositive d))
  end

  include Fraction(M)
  include ExtendOrdered(OCore)

end

(******************************************************************************)
(* The Vector type ************************************************************)
(******************************************************************************)

(** Just as we did for Fractions, we can define (2 dimensional) Vectors as a
    Functor.  As before, this allows us to implement them abstractly once, but
    get multiple implementations "for free": Vector2 (Float),
    Vector2(Fraction(Int)), Vector2(Fraction(BigInt)) and so on.
    *)
module type Vector2 = sig

  module Scalar : Dividable

  include Addable

  (* additional functions that Vectors support that general Addables do not *)
  val dot   : t -> t -> Scalar.t
  val cross : t -> t
  val norm2 : t -> Scalar.t
  val smult : Scalar.t -> t -> t
  val sdiv  : t -> Scalar.t -> t
end



module Vector2 (D : Dividable) = struct
  module Scalar = D

  module Core = struct
    type t = D.t * D.t

    let (+)  (x1,y1) (x2,y2) = D.(x1+x2, y1+y2)
    let (~-) (x,y)           = D.(-x, -y)
    let zero                 = D.(zero, zero)
    let (=)  (x1,y1) (x2,y2) = D.(x1 = x2 && y1 = y2)
  end

  include Core
  include ExtendAddable(Core)

  (** we can define additional functions that aren't part of the general
      addable interface *)
  let dot   (x1,y1) (x2,y2) = D.(x1*x2 + y1*y2)
  let cross (x,y)           = D.(-y,x)
  let norm2 v               = dot v v
  let smult s (x,y)         = D.(s*x,s*y)
  let sdiv  (x,y) s         = D.(x/s,s/y)
end

(** We define a few useful modules from the functors we have built *)

module Rational     = OrdFraction(Int)
module BigRational  = OrdFraction(BigInt)
module Vec2OfFloat  = Vector2 (Float)
module Vec2OfBigRat = Vector2 (BigRational)
module Vec2OfRat    = Vector2 (Rational)

